#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_doctrine.h"
#include "shoki_config.h"
#include "shoki_callback.h"
#include "shoki_fp.h"

extern char *fprint_file;
u_char *pcap_userdata = 0;
char *dumpfile = NULL, *real_outfile = NULL;
int callbacks = 1, fragre = 0, linux_kludge = 0, timeval_kludge = 0;
int rewrite = 0, silent = 0;
int firstsec, lastsec, malsize;
struct bpf_program base_fcode;
FILE *logfile = NULL;

pcap_handler out = pcap_print_pt;

struct hlist hostlist;
struct pfprint pfplist;

extern int nfilters;

int main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL;
	struct sigaction *cleansig;
	struct sigaction *childsig;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "c:C:E:fF:kKL:r:s:S:v")) != EOF)
		switch(op)
		{
			case 'c':
				count = atoi(optarg);
				break;
			case 'C':
				cfile = optarg;
				break;
			case 'E':
				stop = atoi(optarg);
				lastsec = stop;
				break;
			case 'f':
				fragre = 1;
				break;
			case 'k':
				linux_kludge = 1;
				break;
			case 'K':
				timeval_kludge = 1;
				break;
			case 'L':
				currentfile = optarg;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);
				firstsec = start;
				break;
			case 'v':
				verbose = 1;
				break;
		}

	base_filter = get_filter(&argv[optind]);

	cleansig = malloc(sizeof(struct sigaction));
	childsig = malloc(sizeof(struct sigaction));

	childsig->sa_handler = (void *)clean_kids;
	childsig->sa_flags = SA_RESTART & SA_NOCLDSTOP;
	cleansig->sa_handler = (void *)cleanup;
	cleansig->sa_flags = SA_RESTART & SA_NOCLDSTOP;

	(void)sigaction(SIGCHLD, childsig, NULL);
	(void)sigaction(SIGTERM, cleansig, NULL);
	(void)sigaction(SIGINT, cleansig, NULL);

	((cfile == NULL) ? (parse_shoki_config(FP_CONFIG)) : 
		(parse_shoki_config(cfile)));
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	(void)read_passive_fingerprints(fprint_file);
	if(dumpfile == NULL)
		error("Specify a dump file, dammit.");
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	pd = fake_pcap(pd, 0);
	/* Bad developer.  Bad.  No biscuit. */
	if(currentfile != NULL)
	{
		if(*currentfile == '-')
			logfile = stdout;
		else
			logfile = fopen(currentfile, "a");
	}
       	if(logfile == NULL) (void)log_msg(NULL, NULL, NULL, "fp starting");
	(void)read_zdumps(dumpfile, out);
	(void)report();
	if(logfile) (void)fclose(logfile);
	cleanup(0);
	exit(0);
}

int
cleanup(int signo)
{
	if(pd != NULL && pcap_file(pd) == NULL)
	{
		(void)fflush(stdout);
		putc('\n', stderr);
	}
       	if(logfile == NULL) (void)log_msg(NULL, NULL, NULL, "fp shutting down");
	exit(0);
}

void
clean_kids()
{
	waitpid(-1, NULL, WNOHANG);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	const struct timeval *tvp;
	struct ip *ip;
	struct tcphdr *tp;
	struct udphdr *up;
	u_char flags = 0;
	int hlen, ttl, win = 0, df = 0, tos;
	struct hlist *hl;
	struct seqlist *sq;
	struct pfprint *fp;

	(const struct timeval *)tvp = &h->ts;
	if(stop && (tvp->tv_sec > stop)) return;
	if(start && (tvp->tv_sec < start)) return;
	ip = (struct ip *)(sp + sizeof(struct ether_header));
	hlen = (ip->ip_hl * 4) - 20;
	for(hl = &hostlist; hl->next; hl = hl->next)
		if(hl->addr == ip->ip_src.s_addr) goto skip;
	if(!(hl->seqs = (struct seqlist *)calloc(1, sizeof(struct seqlist))))
		error("calloc:  pcap_print_pt");
	if(!(hl->next = (struct hlist *)calloc(1, sizeof(struct hlist))))
		error("calloc:  pcap_print_pt");
	if(!(hl->next->seqs = (struct seqlist *)calloc(1, sizeof(struct seqlist))))
		error("calloc:  pcap_print_pt");
	hl->addr = ip->ip_src.s_addr;
skip:
	ttl = ip->ip_ttl;
	if(ntohs(ip->ip_off) & IP_DF) df = 1;
	tos = ip->ip_tos;
	switch(ip->ip_p)
	{
		case IPPROTO_TCP:
			tp = (struct tcphdr *)(sp + sizeof(struct ether_header) + sizeof(struct ip) + hlen);
			win = ntohs(tp->th_win);
			flags = tp->th_flags;
			if(!(flags & TH_SYN)) break;
			if(flags & TH_ACK) break;
			for(sq = hl->seqs; sq->next; sq = sq->next);
			if(!(sq->next = (struct seqlist *)calloc(1, sizeof(struct seqlist))))
				error("calloc:  pcap_print_pt");
			sq->seq = ntohl(tp->th_seq);
			sq->sec = tvp->tv_sec;
			break;
		case IPPROTO_UDP:
			up = (struct udphdr *)(sp + sizeof(struct ether_header) + sizeof(struct ip) + hlen);
			break;
		default:
			return;
			break;
	}
	hl->index++;
	for(fp = &pfplist; fp->next; fp = fp->next)
	{
		if(ttl > fp->ttl) continue;
		/* Blithely assume nothing's more than 32 hops
		   away. */
		if(ttl < (fp->ttl - 32)) continue;
		if(win < fp->win_low) continue;
		if(win > fp->win_high) continue;
		if(df != fp->df) continue;
		if(tos != fp->tos) continue;
		hl->fps[fp->index]++;
	}
}

void
report()
{
	struct hlist *h;
	struct pfprint *fp;
	char buf[80];
	float f;
	int i;

	for(h = &hostlist; h->next; h = h->next)
	{
		i = 0;
		for(fp = &pfplist; fp->next; fp = fp->next)
		{
			if(h->fps[fp->index] == 0) continue;
			f = ((float)h->fps[fp->index] / (float)h->index) * 100.0;
			snprintf(buf, 80, "%s %s %s", fp->os, fp->version,
				fp->platform);
			(void)log_msg(NULL, NULL, NULL,
				"%-15s     %-35s    %5.0f",
				getname((u_char *)&h->addr), buf, f);
			i++;
		}
		if(i == 0)
			if(verbose) (void)log_msg(NULL, NULL, NULL,
				"%s\tNo matches found",
				getname((u_char *)&h->addr));
	}
}

u_long
get_gcd(u_long foo, u_long bar)
{
	u_long baz;

	if(foo < bar)
	{
		baz = foo;
		foo = bar;
		bar = baz;
	}
	while(bar)
	{
		baz = foo % bar;
		foo = bar;
		bar = baz;
	}
	return(foo);
}

#define WHITE "\t "

void
read_passive_fingerprints(const char *config)
{
	FILE *f = NULL;
	char line[1024];
	char *foo, *baz;
	struct pfprint *fp;
	int l = 0, j = 0;
	
	f = fopen(config, "r");
	if(f == NULL) error("Couldn't read %s", config);
	fp = &pfplist;
	while(fgets(line, sizeof(line), f))
	{
		l++;
		foo = line + strspn(line, WHITE);
		if(*foo == '#' || *foo == '\n' || !*foo) continue;
		foo = strtok(foo, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		if(!(fp->os = (char *)calloc(1, sizeof(*foo))))
			error("calloc:  read_passive_fingerprints");
		fp->os = strdup(foo);
		foo = strtok(NULL, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		fp->version = strdup(foo);
		foo = strtok(NULL, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		if(!(fp->platform = (char *)calloc(1, sizeof(*foo))))
			error("calloc:  read_passive_fingerprints");
		fp->platform = strdup(foo);
		foo = strtok(NULL, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		fp->ttl = atoi(foo);
		foo = strtok(NULL, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		if((baz = rindex(foo, '-')))
		{
			*baz = '\0';
			baz++;
			fp->win_low = atoi(foo);
			fp->win_high = atoi(baz);
		}
		else
		{
			fp->win_low = atoi(foo);
			fp->win_high = atoi(foo);
		}
		foo = strtok(NULL, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		if(strncasecmp(foo, "y", 1) == 0)
			fp->df = 1;
		else
			fp->df = 0;
		foo = strtok(NULL, WHITE);
		if(foo == NULL || !*foo || *foo == '#')
			error("Bad fingerprint in %s, line %d", config, l);
		fp->tos = atoi(foo);
		j++;
		fp->index = j;
		if(!(fp->next = (struct pfprint *)calloc(1, sizeof(struct pfprint))))
			error("calloc:  read_passive_fingerprints");
		fp = fp->next;
	}
}
