#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_doctrine.h"
#include "shoki_callback.h"

extern int fcount_pt[MAX_FILTERS], nfilters;
extern int d_pt[MAX_FILTERS], t_pt[MAX_FILTERS], q_pt[MAX_FILTERS];
extern int h_pt[MAX_FILTERS], sev_pt[MAX_FILTERS];
extern char **filter_pt, **hex_pt, **call_args_pt, **id_pt, **type_pt;
extern regex_t **regex_pt;
extern int callback_pt[MAX_FILTERS], rn_pt[MAX_FILTERS];
extern struct bpf_program **fcode_pt, base_fcode;
extern char email_rcpt[256];
extern FILE *logfile;
extern char *shoki_log_name;
extern char *shoki_type;
extern int fragre, silent, verbose;

char outbuf[OUT_BUFSIZE];

#define NEW_LIST(ptr, type) \
	if(!(ptr = (struct type *)calloc(1, sizeof(struct type)))) \
		error("NEW_LIST:  calloc"); ptr->prev = NULL
#define APPEND_LIST(ptr, tmp, type) \
	if(!(tmp = (struct type *)calloc(1, sizeof(struct type)))) \
		error("NEW_LIST:  calloc"); ptr->next = tmp; \
		tmp->prev = ptr

#define TH_HASH_SIZE	907
int th_hash[TH_HASH_SIZE];
int last_sec[MAX_FILTERS], flagged_sec[MAX_FILTERS];
u_int32_t runtime;

#ifndef SHOKI_LOG_NAME
#define SHOKI_LOG_NAME	"shoki"
#endif

struct srclist srclist[TH_HASH_SIZE];
struct flist *flist = NULL;

static struct { const char *name; int option; } callback_list[] =
{
	{ "NULL",	CALLBACK_NULL },
	{ "LOG",	CALLBACK_LOG },
	{ "MAIL",	CALLBACK_MAIL },
	{ "SEARCH",	CALLBACK_SEARCH },
	{ "REMOTE",	CALLBACK_REMOTE },
	{ "MESHUGA_TSEQ",	CALLBACK_MESHUGA_TSEQ },
	{ "MESHUGA_T1",	CALLBACK_MESHUGA_T1 },
	{ "MESHUGA_T2",	CALLBACK_MESHUGA_T2 },
	{ "MESHUGA_T3",	CALLBACK_MESHUGA_T3 },
	{ "MESHUGA_T4",	CALLBACK_MESHUGA_T4 },
	{ "MESHUGA_T5",	CALLBACK_MESHUGA_T5 },
	{ "MESHUGA_T6",	CALLBACK_MESHUGA_T6 },
	{ "MESHUGA_T7",	CALLBACK_MESHUGA_T7 },
	{ "MESHUGA_PU",	CALLBACK_MESHUGA_PU },
	{ "THRESHOLD",	CALLBACK_THRESHOLD },
	{ "HOST_SCAN",	CALLBACK_HOST_SCAN },
	{ "PORT_SCAN",	CALLBACK_PORT_SCAN },
	{ NULL,		0 }
};

extern int pf_outformat[];
extern char *pf_outdelim[];
extern int pf_type[];

int
callback_handler(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h, int output)
{
	switch(filter->callback)
	{
		case CALLBACK_LOG:
			return(callback_log(sp, filter, len, h));
			break;
		case CALLBACK_SEARCH:
			return(callback_search(sp, filter, len, h));
			break;
		case CALLBACK_REMOTE:
			return(callback_remote(sp, filter, len, h));
			break;
		case CALLBACK_THRESHOLD:
			return(callback_threshold(sp, filter, len, h));
			break;
		case CALLBACK_HOST_SCAN:
			return(callback_host_scan(sp, filter, len, h));
			break;
		case CALLBACK_PORT_SCAN:
			return(callback_port_scan(sp, filter, len, h));
			break;
		default:
			return(-1);
			error("Unknown callback called.");
			break;
	}
}

int
callback_threshold(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h)
{
	int t, foo, bar, gap, t0, i;
	char baz[512];

	i = filter->index;
	if((filter->hex != NULL) &&
		(regexsearch(sp, len, filter->regex, strlen(filter->hex)) == 0))
		return(0);
	t = h->ts.tv_sec;
	th_hash[t % TH_HASH_SIZE]++;
	gap = t - last_sec[i];
	if((gap > 1) && (last_sec[i] != 0))
		for(foo = (last_sec[i] + 1); foo < t; foo++)
			th_hash[foo % TH_HASH_SIZE] = 0;
	t0 = t - (filter->d - 1);
	th_hash[(t0 - 1) % TH_HASH_SIZE] = 0;
	bar = 0;
	for(foo = t0; foo <= t; foo++)
		bar += th_hash[foo % TH_HASH_SIZE];
	last_sec[i] = t;
	if(t <= (flagged_sec[i] + filter->q)) return(0);
	if(bar > filter->t)
	{
		flagged_sec[i] = t;
		sp += sizeof(struct ether_header);
		snprintf(baz, 512, "%s (%d:%d)", filter->call_args, bar, filter->d);
		(void)log_packet(sp, h, pf_outformat, baz, filter->id, sp);
		return(1);
	}
	return(0);
}

int
callback_host_scan(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h)
{
	struct srclist *s;
	struct ether_header *ep;
	u_short ether_type;
	struct ip *ip;
	u_int hlen, l;
	int i, j, t, foo, bar, gap, t0;
	char baz[512];

	i = filter->index;
	ep = (struct ether_header *)sp;
	sp += sizeof(struct ether_header);
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP) return(0);
	ip = (struct ip *)sp;
	hlen = (ip->ip_hl * 4) - 20;
	l = ntohs(ip->ip_len) - hlen;
	t = h->ts.tv_sec;
	foo = t % TH_HASH_SIZE;
	(void)hs_add(ip->ip_src.s_addr, ip->ip_dst.s_addr, foo, 1, -1);
	if(t <= (flagged_sec[i] + filter->q)) return(0);
	t0 = t - (filter->d - 1);
	gap = t - last_sec[i];
	if((gap > 1) && (last_sec[i] != 0))
		for(bar = (last_sec[i] + 1); bar < t; bar++)
			for(s = &srclist[bar % TH_HASH_SIZE]; s->next; s = s->next)
				{
					s->total = 0;
					memset(s->dstlist, '\0', sizeof(struct dstlist));
					memset(s->ports, '\0', sizeof(struct plist));
				}
	bar = 0;
	for(j = t0; j <= t; j++)
		for(s = &srclist[j % TH_HASH_SIZE]; s->next; s = s->next)
			if(s->total > 0) bar++;
	if(bar > filter->t)
	{
		if(filter->h == 1)
		{
			 return(permute_hosts(ip->ip_src.s_addr, sp, h, pf_outformat, filter, t0, t));
		}
		else
		{
			flagged_sec[i] = t;
			snprintf(baz, 512, "%s (%d:%d)", filter->call_args, bar, filter->d);
			(void)log_packet(sp, h, pf_outformat, baz, filter->id, sp);
			return(1);
		}
	}
	last_sec[i] = t;
	return(0);
}

int
callback_port_scan(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h)
{
	struct srclist *s;
	struct ether_header *ep;
	u_short ether_type;
	const u_char *cp;
	struct ip *ip;
	struct tcphdr *tp;
	struct udphdr *up;
	char baz[512];
	u_int hlen, l, off;
	u_short dport = 0;
	int i, t, foo, bar, j, t0, gap;
	struct plist *p;

	i = filter->index;
	ep = (struct ether_header *)sp;
	sp += sizeof(struct ether_header);
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP) return(0);
	ip = (struct ip *)sp;
	hlen = (ip->ip_hl * 4) - 20;
	l = ntohs(ip->ip_len) - hlen;
	off = ntohs(ip->ip_off);
	t = h->ts.tv_sec;
	foo = t % TH_HASH_SIZE;
	/* We need the layer 3 headers */
	if((off & 0x1fff) == 0)
	{
		cp = (const u_char *)ip + hlen;
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
				if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
					error("calloc:  callback_port_scan");
				memcpy(tp, (sp + sizeof(struct ip)), sizeof(struct tcphdr));
				dport = ntohs(tp->th_dport);
				free(tp);
				break;
			case IPPROTO_UDP:
				if(!(up = (struct udphdr *)calloc(1, sizeof(struct udphdr))))
					error("calloc:  callback_port_scan");
				memcpy(up, (sp + sizeof(struct ip)), sizeof(struct udphdr));
				dport = ntohs(up->uh_dport);
				free(up);
				break;
		}
		if(dport == 0) return(0);
		(void)hs_add(ip->ip_src.s_addr, ip->ip_dst.s_addr, foo, 0, dport);
		if(t <= (flagged_sec[i] + filter->q)) return(0);
		t0 = t - (filter->d - 1);
		gap = t - last_sec[i];
		if((gap > 1) && (last_sec[i] != 0))
			for(bar = (last_sec[i] + 1); bar < t; bar++)
				for(s = &srclist[bar % TH_HASH_SIZE]; s->next; s = s->next)
				{
					s->total = 0;
					memset(s->dstlist, '\0', sizeof(struct dstlist));
					memset(s->ports, '\0', sizeof(struct plist));
				}
		bar = 0;
		for(j = t0; j <= t; j++)
			for(s = &srclist[j % TH_HASH_SIZE]; s->next; s = s->next)
			{
				bar = 0;
				for(p = s->ports; p->next; p = p->next)
					if(p->total > 0) bar++;
				if(bar > filter->t)
				{
					flagged_sec[i] = t;
					snprintf(baz, 512, "%s (%d:%d)", filter->call_args, bar, filter->d);
					(void)log_packet(sp, h, pf_outformat, baz, filter->id, sp);
					return(1);
				}
				last_sec[i] = t;
			}
	}
	return(0);
}

int
hs_add(u_int32_t saddr, u_int32_t daddr, int i, int count, int port)
{
	struct srclist *s;
	struct dstlist *d;
	struct plist *p;
	int j = 0, k = 0, l = 0;;

	for(s = &srclist[i]; s->next && j == 0; s = s->next)
		if(s->src == saddr)
		{
			k = 0;
			for(d = s->dstlist; d->next && k == 0; d = d->next)
				if(d->dst == daddr)
				{
					d->count += count;
					k++;
				}
			if(k == 0)
			{
				d->count = count;
				s->total++;
				d->dst = daddr;
				if(!(d->next = (struct dstlist *)calloc(1, sizeof(struct dstlist))))
					error("calloc:  hs_add");
			}
			if((port > -1) && (port < 65536))
			{
				k = 0;
				for(p = s->ports; p->next && k == 0; p = p->next)
					if(p->port == port)
					{
						l = 0;
						for(d = p->dstlist; d->next && l == 0; d = d->next)
						{
							d->count++;
							l++;
						}
						if(l == 0)
						{
							d->count = 1;
							p->total++;
							d->dst = daddr;
							if(!(d->next = (struct dstlist *)calloc(1, sizeof(struct dstlist))))
							error("calloc:  hs_add");
						}
						k++;
					}
					if(k == 0)
					{
						p->port = port;
						p->total = 1;
						if(!(p->dstlist = (struct dstlist *)calloc(1, sizeof(struct dstlist))))
							error("calloc:  hs_add");
						if(!(p->next = (struct plist *)calloc(1, sizeof(struct plist))))
							error("calloc:  hs_add");
					}
			}
			j++;
		}
	if(j == 0)
	{
		s->src = saddr;
		s->total = 1;
		if(!(s->dstlist = (struct dstlist *)calloc(1, sizeof(struct dstlist))))
			error("calloc:  hs_add");
		s->dstlist->dst = daddr;
		s->dstlist->count = count;
		if(!(s->dstlist->next = (struct dstlist *)calloc(1, sizeof(struct dstlist))))
			error("calloc:  hs_add");
		if(!(s->next =
			(struct srclist *)calloc(1, sizeof(struct srclist))))
			error("calloc:  hs_add");
		if(!(s->ports = (struct plist *)calloc(1, sizeof(struct plist))))
			error("calloc:  hs_add");
		if((port > -1) && (port < 65536))
		{
			s->ports->port = port;
			s->ports->total = 1;
			if(!(s->ports->dstlist = (struct dstlist *)calloc(1, sizeof(struct dstlist))))
				error("calloc:  hs_add");
			if(!(s->ports->next = (struct plist *)calloc(1, sizeof(struct plist))))
				error("calloc:  hs_add");
		}
	}
	return(1);
}

int
callback_remote(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h)
{
	struct ether_header *ep;
	u_short ether_type;
	const u_char *cp, *rsp;
	const struct ip *ip;
	u_int hlen, l, off;
	int i, st = 0;

	i = filter->index;
	ep = (struct ether_header *)sp;
	sp += sizeof(struct ether_header);
	rsp = sp;
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP) return(0);
	ip = (const struct ip *)sp;
	hlen = (ip->ip_hl * 4) - 20;
	l = ntohs(ip->ip_len) - hlen;
	off = ntohs(ip->ip_off);
	if((off & 0x1fff) == 0)
	{
		cp = (const u_char *)ip + hlen;
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
			case IPPROTO_UDP:
			case IPPROTO_ICMP:
				packet_print(cp, l, (const u_char *)ip, ip->ip_p, h, filter, rsp);
				st = 1;
				break;
			default:
				break;
		}
	}
	return(st);
}

int
callback_log(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h)
{
	struct ether_header *ep;
	u_short ether_type;
	u_char *cp, *rsp;
	struct ip *ip;
	u_int hlen, l, off;
	int st = 0;

	if(!(ep = (struct ether_header *)calloc(1, sizeof(struct ether_header))))
		error("calloc:  callback_log");
	memcpy(ep, sp, sizeof(struct ether_header));
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP)
	{
		(void)log_msg(h, NULL, NULL, "Non-IP packet");
		free(ep);
		return(0);
	}
	sp += sizeof(struct ether_header);
	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  callback_log");
	memcpy((u_char *)ip, sp, sizeof(struct ip));
	hlen = (ip->ip_hl * 4) - 20;
	if(h->caplen < ntohs(ip->ip_len))
	{
		l = h->caplen;
	}
	else
	{
		l = ntohs(ip->ip_len);
	}
	if(!(rsp = (u_char *)calloc(1, l)))
		error("calloc:  callback_log");
	memcpy((u_char *)rsp, (u_char *)sp, l);
	l = ntohs(ip->ip_len) - hlen;
	if(!(cp = (u_char *)calloc(1, l)))
		error("calloc:  callback_log");
	off = ntohs(ip->ip_off);
	if((off & 0x1fff) == 0)
	{
		/* This is kinda stupid.  The packet has been broken up
		   into a pointer to the layer 2 headers (always IP)
		   and a pointer to the layer 3 headers.  Anything between
		   (i.e., IP options) are lost at this point.   -spb
		*/
		memcpy((u_char *)cp, (u_char *)(sp + sizeof(struct ip) + hlen), l);
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
			case IPPROTO_UDP:
			case IPPROTO_ICMP:
				packet_print(cp, l, (const u_char *)ip, ip->ip_p, h, filter, rsp);
				st = 1;
				break;
			default:
				break;
		}
	}
	free(ep);
	free(ip);
	free(cp);
	free(rsp);
	return(st);
}

void
remote_packet_print(const u_char *bp, u_int length, const u_char *bp2, u_char proto, const struct pcap_pkthdr *h, int filt)
{
}

void
packet_print(const u_char *bp, u_int length, const u_char *bp2, u_char proto, const struct pcap_pkthdr *h, struct filter *filter, const u_char *rsp)
{
	struct tcphdr *tp = NULL;
	struct udphdr *up = NULL;
	struct icmp *dp = NULL;
	struct ip *ip = NULL;
	char *sp = NULL;
	u_int off;

	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  packet_print");
	memcpy(ip, bp2, sizeof(struct ip));
	off = ntohs(ip->ip_off);
	if((off & 0x1fff) == 0)
	{
		switch(ip->ip_p)
		{
			case IPPROTO_TCP:
				if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
					error("calloc:  packet_print");
				memcpy(tp, bp, sizeof(struct tcphdr));
				if(!(sp = (char *)calloc(1, (sizeof(struct ip) + sizeof(struct tcphdr)))))
					error("calloc:  packet_print");
				memcpy(sp, (char *)ip, sizeof(struct ip));
				memcpy((char *)(sp + sizeof(struct ip)), (char *)tp, sizeof(struct tcphdr));
				free(tp);
				break;
			case IPPROTO_UDP:
				if(!(up = (struct udphdr *)calloc(1, sizeof(struct udphdr))))
					error("calloc:  packet_print");
				memcpy(up, bp, sizeof(struct udphdr));
				if(!(sp = (char *)calloc(1, (sizeof(struct ip) + sizeof(struct udphdr)))))
					error("calloc:  packet_print");
				memcpy(sp, (char *)ip, sizeof(struct ip));
				memcpy((char *)(sp + sizeof(struct ip)), (char *)up, sizeof(struct udphdr));
				free(up);
				break;
			case IPPROTO_ICMP:
				if(!(dp = (struct icmp *)calloc(1, sizeof(struct icmp))))
					error("calloc:  packet_print");
				memcpy(dp, bp, sizeof(struct icmp));
				if(!(sp = (char *)calloc(1, (sizeof(struct ip) + sizeof(struct icmp)))))
					error("calloc:  packet_print");
				memcpy(sp, (char *)ip, sizeof(struct ip));
				memcpy((char *)(sp + sizeof(struct ip)), (char *)dp, sizeof(struct icmp));
				free(dp);
				break;
		}
	}
	(void)log_packet(sp, h, pf_outformat, filter->call_args, filter->id, rsp);
	free(ip);
	free(sp);
}

char *
packet_output(const u_char *sp, const struct pcap_pkthdr *h, int outformat[], char *comment, char *id, const u_char *rsp)
{
	struct tcphdr *tp = NULL;
	struct udphdr *up = NULL;
	struct icmp *dp = NULL;
	struct ip *ip = NULL;
	struct timeval *tvp;
	u_char flags = 0, icmp_type = 0;
	u_int hlen = 0, p2;
	u_short sport = 0, dport = 0, win = 0, *pp;
	u_int32_t seq = 0, ack = 0;
	u_int8_t proto = 0;
	int i, length = 0, tl = 0, num = 0, k = 0, m = 0, n = 0;
	char buf[OUT_BUFSIZE], log[OUT_BUFSIZE], tmp[OUT_BUFSIZE];

	if(runtime == 0) runtime = time(NULL);
	if(!(tvp = (struct timeval *)calloc(1, sizeof(struct timeval))))
		error("calloc:  packet_output");
	memcpy(tvp, &h->ts, sizeof(struct timeval));
	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  packet_output");
	if(sp == NULL) goto print;
	memcpy(ip, sp, sizeof(struct ip));
	length = ntohs(ip->ip_len);
	tl = length;
	length -= ((ip->ip_hl * 4) - 20);
	proto = ip->ip_p;
	
	switch(proto)
	{
		case IPPROTO_TCP:
			if(!(tp = (struct tcphdr *)calloc(1, sizeof(struct tcphdr))))
				error("calloc:  packet_output");
			memcpy(tp, (sp + sizeof(struct ip)), sizeof(struct tcphdr));
			sport = ntohs(tp->th_sport);
			dport = ntohs(tp->th_dport);
			seq = ntohl(tp->th_seq);
			ack = ntohl(tp->th_ack);
			win = ntohs(tp->th_win);
			flags = tp->th_flags;
			icmp_type = 0;
			hlen = tp->th_off * 4;
			length -= hlen;
			free(tp);
			break;
		case IPPROTO_UDP:
			if(!(up = (struct udphdr *)calloc(1, sizeof(struct udphdr))))
				error("calloc:  packet_output");
			memcpy(up, (sp + sizeof(struct ip)), sizeof(struct udphdr));
			sport = ntohs(up->uh_sport);
			dport = ntohs(up->uh_dport);
			seq = 0;
			ack = 0;
			flags = 0;
			icmp_type = 0;
			free(up);
			break;
		case IPPROTO_ICMP:
			if(!(dp = (struct icmp *)calloc(1, sizeof(struct icmp))))
				error("calloc:  packet_output");
			memcpy(dp, (sp + sizeof(struct ip)), sizeof(struct icmp));
			icmp_type = dp->icmp_type;
			seq = 0;
			ack = 0;
			sport = 0;
			dport = 0;
			flags = 0;
			free(dp);
			break;
		default:
			break;
	}
print:

	for(i = 0; pf_type[i] != 0; i++)
	{
		if(pf_type[i] == PF_TYPE_DELIM)
		{
			if((strlen(log) + strlen(pf_outdelim[m])) > OUT_BUFSIZE)
				error("Output buffer length exceeded in packet_output");
			(((n == 0) && (m == 0)) ? strcpy(log, pf_outdelim[m]) : strcat(log, pf_outdelim[m]));
			m++;
			continue;
		}
		buf[0] = '\0';
		switch(outformat[n])
		{
			case PF_RUNTIME:
				sprintf(buf, "%u", (u_int32_t)runtime);
				break;
			case PF_ID:
				sprintf(buf, "%s", id);
				break;
			case PF_NAME:
				sprintf(buf, "%s", comment);
				break;
			case PF_TIME:
				sprintf(buf, "%u.%06u", (u_int32_t)tvp->tv_sec,
					(u_int32_t)tvp->tv_usec);
				break;
			case PF_SEC:
				sprintf(buf, "%u", (u_int32_t)tvp->tv_sec);
				break;
			case PF_USEC:
				sprintf(buf, "%06u", (u_int32_t)tvp->tv_usec);
				break;
			case PF_SRC:
				sprintf(buf, "%s", ipaddr_string(&ip->ip_src));
				break;
			case PF_DST:
				sprintf(buf, "%s",
						ipaddr_string(&ip->ip_dst));
				break;
			case PF_SPORT:
				if((proto == IPPROTO_TCP) ||
					(proto == IPPROTO_UDP))
				{
					sprintf(buf, "%u", sport);
				}
				else
				{
					sprintf(buf, "0");
				}
				break;
			case PF_DPORT:
				if((proto == IPPROTO_TCP) ||
					(proto == IPPROTO_UDP))
				{
					sprintf(buf, "%u", dport);
				}
				else
				{
					sprintf(buf, "0");
				}
				break;
			case PF_PROTO:
				sprintf(buf, "%d", proto);
				break;
			case PF_SEQ:
				if(seq)
					{sprintf(buf, "%u", seq);}
				else
					{sprintf(buf, "0");}
				break;
			case PF_LEN:
				if(tl) {sprintf(buf, "%d", tl);}
				else {sprintf(buf, "0");}
				break;
			case PF_SYN:
				if(flags) ((flags & TH_SYN) ? sprintf(buf, "1") : sprintf(buf, "0"));
				break;
			case PF_FIN:
				if(flags) ((flags & TH_FIN) ? sprintf(buf, "1") : sprintf(buf, "0"));
				break;
			case PF_RST:
				if(flags) ((flags & TH_RST) ? sprintf(buf, "1") : sprintf(buf, "0"));
				break;
			case PF_PUSH:
				if(flags) ((flags & TH_PUSH) ? sprintf(buf, "1") : sprintf(buf, "0"));
				break;
			case PF_ACK:
				if(ack) {sprintf(buf, "%u", ack);}
				else {sprintf(buf, "0");}
				break;
			case PF_ICMP_TYPE:
				if(icmp_type) {sprintf(buf, "%d", icmp_type);}
				else {sprintf(buf, "0");}
				break;
			case PF_WIN:
				if(win) {sprintf(buf, "%d", win);}
				else {sprintf(buf, "0");}
				break;
			case PF_FLAGS:
				if(flags)
				{
					if(flags & TH_FIN) strcat(buf, "F");
					if(flags & TH_SYN) strcat(buf, "S");
					if(flags & TH_RST) strcat(buf, "R");
					if(flags & TH_PUSH) strcat(buf, "P");
					if(flags & TH_ACK) strcat(buf, "A");
					if(flags & TH_URG) strcat(buf, "U");
				}
				break;
			case PF_HEX:
				pp = (u_short *)rsp;
				k = 0;
				if(h->caplen < ntohs(ip->ip_len))
				{
					num = h->caplen / sizeof(u_short);
				}
				else
				{
					num = ntohs(ip->ip_len) / sizeof(u_short);
				}
				if((long)rsp & 1)
				{
					while(--num >= 0)
					{
						p2 = *pp++;
						if(strlen(buf) > (OUT_BUFSIZE - 4))
							error("Output buffer length exceeded in packet_output");
						sprintf(tmp, "%02x%02x", p2, *pp++);
						strcat(buf, tmp);
					}
				}
				else
				{
					while(--num >= 0)
					{
						if(strlen(buf) > (OUT_BUFSIZE - 4))
							error("Output buffer length exceeded in packet_output");
						sprintf(tmp, "%04x", ntohs(*pp++));
						strcat(buf, tmp);
					}
				}
				if(h->caplen & 1)
				{
					if(strlen(buf) > (OUT_BUFSIZE - 2))
						error("Output buffer length exceeded in packet_output");
					sprintf(tmp, "%02x", *(u_char *)pp);
					strcat(buf, tmp);
				}
				break;
			case PF_TOS:
				if(ip->ip_tos) {sprintf(buf, "%d", ip->ip_tos);}
				else {sprintf(buf, "0");}
				break;
			case PF_TTL:
				if(ip->ip_ttl) {sprintf(buf, "%d", ip->ip_ttl);}
				else {sprintf(buf, "0");}
				break;
			case PF_IPID:
				if(ip->ip_id) {sprintf(buf, "%u", ntohs(ip->ip_id));}
				else {sprintf(buf, "0");}
				break;
			case PF_SENSOR:
				sprintf(buf, "%s", shoki_type);
				break;
			case PF_ALERT:
				sprintf(buf, "0");
				break;
			default:
				break;
		}
		if(((strlen(log) + strlen(buf)) * sizeof(char)) > OUT_BUFSIZE)
			error("Output buffer length exceeded in packet_output");
		(((n == 0) && (m == 0)) ? strcpy(log, buf) : strcat(log, buf));
		n++;
	}
	if((strlen(log) * sizeof(char)) > (OUT_BUFSIZE))
		error("Output buffer length exceeded in packet_output");
	strncpy(outbuf, log, OUT_BUFSIZE);
	free(ip);
	if(tvp) free(tvp);
	return(outbuf);
}

int
callback_search(const u_char *sp, struct filter *filter, int len, const struct pcap_pkthdr *h)
{
	if(strlen(filter->hex) == 0) return(0);
	if(regexsearch(sp, len, filter->regex, strlen(filter->hex)) == 0)
		return(0);
	return(callback_log(sp, filter, len, h));
}

char *
getname(const u_char *ap)
{
	u_int32_t addr;
	addr = *(const u_int32_t *)ap;
	switch((long)ap & 3)
	{
		case 0:
			addr = *(u_int32_t *)ap;
			break;
		case 2:
			addr = ((u_int32_t)*(u_short *)(ap + 2) << 16) |
				(u_int32_t)*(u_short *)ap;
			break;
		default:
			addr = ((u_int32_t)ap[3] << 24) |
				((u_int32_t)ap[2] << 16) |
				((u_int32_t)ap[1] << 8) |
				(u_int32_t)ap[0];
			break;
	}
	return(intoa(addr));
}

char *
intoa(u_int32_t addr)
{
	register char *cp;
	register u_int byte;
	register int n;
	static char buf[sizeof(".xxx.xxx.xxx.xxx")];

	addr = ntohl(addr);
	cp = &buf[sizeof buf];
	*--cp = '\0';

	n = 4;
	do
	{
		byte = addr & 0xff;
		*--cp = byte % 10 + '0';
		byte /= 10;
		if(byte > 0)
		{
			*--cp = byte % 10 + '0';
			byte /= 10;
			if(byte > 0)
				*--cp = byte + '0';
		}
		*--cp = '.';
		addr >>= 8;
	} while (--n > 0);
	return cp + 1;
}


int
get_callback(char *foo)
{
	int i = 0;
	
	while(callback_list[i].name)
	{
		if(strncmp(foo, callback_list[i].name, strlen(foo)) == 0)
			return(callback_list[i].option);
		i++;
	}
	error("Unknown callback %s invoked.\n", foo);
	return(0);
}

int
permute_hosts(u_int32_t saddr, const u_char *sp, const struct pcap_pkthdr *h, int outformat[], struct filter *filter, int t1, int t2)
{
	struct srclist *s;
	int i, rule;
	char baz[512];

	rule = filter->index;
	for(i = t1; i <= t2; i++)
	{
		for(s = &srclist[i % TH_HASH_SIZE]; s->next; s = s->next)
		{
			if(s->src == saddr)
				if(s->total > filter->t)
				{
					flagged_sec[rule] = t2;
					snprintf(baz, 512, "%s (%d)", filter->call_args, s->total);
					(void)log_packet(sp, h, outformat, baz, filter->id, sp);
					return(1);
				}
		}
	}
	return(0);
}

void
log_msg(const struct pcap_pkthdr *h, const u_char *sp, const u_char *rsp, char *fmt, ...)
{
	va_list va;
	char *foo;
	char buf[512];

	if(silent) return;
	va_start(va, fmt);
	vsnprintf(buf, sizeof(buf), fmt, va);
	va_end(va);
	if(&h->ts)
		foo = packet_output(sp, h, pf_outformat, buf, NULL, rsp);
	else
		foo = strdup(buf);
	if(logfile != NULL)
	{
		fprintf(logfile, "%s\n", foo);
	}
	else
	{
		(void)openlog(shoki_log_name, LOG_PID, SHOKI_LOG_SVC);
		syslog(SHOKI_LOG_LVL, "%s", foo);
		(void)closelog();
	}
}

void
log_packet(const u_char *sp, const struct pcap_pkthdr *h, int outformat[], char *comment, char *id, const u_char *rsp)
{
	char *foo;

	if(silent) return;
	foo = packet_output(sp, h, outformat, comment, id, rsp);
	if(logfile)
	{
		fprintf(logfile, "%s\n", foo);
	}
	else
	{
		(void)openlog(shoki_log_name, LOG_PID, SHOKI_LOG_SVC);
		syslog(SHOKI_LOG_LVL, "%s", foo);
		(void)closelog();
	}
}



/* All linked lists, all the time.  There are faster ways.  -spb */
int
check_frag(const struct pcap_pkthdr *h, const u_char *sp)
{
	struct ether_header *ep;
	u_short ether_type;
	struct ip *ip;
	u_char *cp;
	u_int hlen, l, off, id, proto, tos;
	u_int32_t saddr, daddr;
	int i, f0, f1, tl, o, fs, ls;
	struct flist *f, *f2;
	struct frags *r, *r2;

	if(fragre == 0) return(0);
	if(!(ep = (struct ether_header *)calloc(1, sizeof(struct
		ether_header))))
		error("calloc:  check_frag");
	memcpy(ep, sp, sizeof(struct ether_header));
	ether_type = ntohs(ep->ether_type);
	if(ether_type != ETHERTYPE_IP)
	{
		(void)log_msg(h, NULL, NULL, "Non-IP packet");
		free(ep);
		return(0);
	}
	if(!(ip = (struct ip *)calloc(1, sizeof(struct ip))))
		error("calloc:  check_frag");
	memcpy((u_char *)ip, (sp + sizeof(struct ether_header)),
		sizeof(struct ip));
	off = ntohs(ip->ip_off);
	if(!(off & 0x3fff))
	{
		free(ip);
		free(ep);
		return(0);
	}
	hlen = (ip->ip_hl * 4);
	l = ntohs(ip->ip_len) - hlen;
	if(!(cp = (u_char *)calloc(1, l)))
		error("calloc:  check_frag");
	id = ntohs(ip->ip_id);
	proto = ip->ip_p;
	tos = ip->ip_tos;
	saddr = ip->ip_src.s_addr;
	daddr = ip->ip_dst.s_addr;
	if(flist == NULL)
		NEW_LIST(flist, flist);
	for(f = flist; f->next; f = f->next)
	{
		if((f->id == id) && (f->p == proto) && (f->tos == tos)
			&& (f->src == saddr) && (f->dst == daddr))
		{
			tl = 0;
			f0 = 65536;
			f1 = 0;
			ls = 0;
			fs = 0;
			if(f->frags == NULL)
				NEW_LIST(f->frags, frags);
			for(r = f->frags; r->next; r = r->next)
			{
				if(r->len != 0)
				{
					tl += r->len;
					o = ((r->off & 0x1fff) * 8);
					if(o < f0) f0 = o;
					if((o + l) > f1) f1 = (o + l);
					if((r->off & IP_MF) != 0) ls++;
					if(o == 0) fs++;
				}
			}
			APPEND_LIST(r, r2, frags);
			if(!(r->h = (struct pcap_pkthdr *)calloc(1,
				sizeof(struct pcap_pkthdr))))
				error("calloc:  check_frag");
			if(!(r->p = (u_char *)calloc(1, (l + hlen +
				sizeof(struct ether_header)))))
				error("calloc:  check_frag");
			r->len = l;
			r->off = off;
			tl += l;
			o = ((r->off & 0x1fff) * 8);
			if(o < f0) f0 = o;
			if((o + l) > f1) f1 = (o + l);
			if((off & IP_MF) != 0) ls++;
			if((off & 0x1fff) == 0) fs++;
			memcpy(r->h, h, sizeof(struct pcap_pkthdr));
			memcpy(r->p, sp, (l + hlen +
				sizeof(struct ether_header)));
			i++;
			if(((f1 - f0) == tl) && (ls != 0) && (fs != 0))
			{
				(void)frag_reassemble(f->id);
				f->done = 1;
				if(f->prev != NULL)
				{
					f->prev->next = f->next;
					f->next->prev = f->prev;
				}
				else
				{
					flist = f->next;
					flist->prev = NULL;
				}
				for(r = f->frags; (f != flist) &&
					r->next; r = r->next)
				{
					if(r->prev != NULL)
					{
						free(r->p);
						free(r->h);
						free(r->prev);
					}
				}
				if(f == flist)
				{
					free(r->p);
					free(r->h);
				}
				free(r->next);
				free(r);
				f->next = NULL;
				f->prev = NULL;
				free(f);
				f = NULL;
			}
			goto done;
		}
	}
	f->id = id;
	f->p = proto;
	f->tos = tos;
	f->src = saddr;
	f->dst = daddr;
	f->done = 0;
	NEW_LIST(f->frags, frags);
	r = f->frags;
	if(!(r->h = (struct pcap_pkthdr *)calloc(1,
		sizeof(struct pcap_pkthdr))))
		error("calloc:  check_frag");
	if(!(r->p = (u_char *)calloc(1, (l + hlen +
		sizeof(struct ether_header)))))
		error("calloc:  check_frag");
	memcpy(r->h, h, sizeof(struct pcap_pkthdr));
	memcpy(r->p, sp, (l + hlen + sizeof(struct ether_header)));
	f->frags->off = off;
	f->frags->len = l;
	APPEND_LIST(f->frags, r2, frags);
	APPEND_LIST(f, f2, flist);
done:
	free(cp);
	free(ep);
	free(ip);
	return(1);
}

int
frag_reassemble(u_short id)
{
	struct flist *f;
	struct frags *r;
	struct pcap_pkthdr *h = NULL;
	u_char *sp, *cp;
	struct ip *ip = NULL, *ip2;
	u_int hlen, l, rh = 0;
	int o, f0 = 0, f1 = 0, tl = 0, j;
	char *p;

	if(fragre == 0) return(0);
	if(flist == NULL) return(0);
	if(!(h = (struct pcap_pkthdr *)calloc(1, sizeof(struct pcap_pkthdr))))
		error("calloc:  frag_reassemble");
	for(f = flist; f->next; f = f->next)
	{
		if(((id != 0) && (f->id == id)) || (id == 0))
		{
			if(!(sp = (u_char *)calloc(1, 65536)))
				error("calloc:  frag_reassemble");
			if(!(cp = (u_char *)calloc(1, 65536)))
				error("calloc:  frag_reassemble");
			f0 = 65536;
			f1 = 0;
			tl = 0;
			for(r = f->frags; r->next; r = r->next)
			{
				ip = (struct ip *)(r->p + sizeof(struct ether_header));
				hlen = (ip->ip_hl * 4);
				l = ntohs(ip->ip_len) - hlen;
				o = ((r->off & 0x1fff) * 8);
				if(o < f0) f0 = o;
				if((o + l) > f1) f1 = (o + l);
				tl += l;
				if(o == 0)
				{
					rh = sizeof(struct ether_header) + hlen;
					memcpy(sp, r->p, rh);
					h = r->h;
				}
				/* Timestamp in log will correspond to the
				   last frag received.  The linked list still
				   contains the pcap_pkthdr of each frag, so
				   the timestamp of each frag could be logged.
							-spb
				*/
				if(r->h->ts.tv_sec > h->ts.tv_sec)
				{
					h->ts.tv_sec = r->h->ts.tv_sec;
					h->ts.tv_usec = r->h->ts.tv_usec;
				}
				else if(r->h->ts.tv_sec == h->ts.tv_sec)
				{
					if(r->h->ts.tv_usec > h->ts.tv_usec)
					{
						h->ts.tv_sec = r->h->ts.tv_sec;
						h->ts.tv_usec = r->h->ts.tv_usec;
					}
				}
				if((o + l) > 65535)
					(void)log_msg(h, (sp + sizeof(struct ether_header)), (sp + sizeof(struct ether_header)), "Invalid frag (offset %d + len %d > 65535)", o, l);
				else
				{
					/* This is a mean old ugly way to
					   handle it, but we can detect
					   overlapping frags this way.  -spb
					*/
					p = (cp + o);
					j = 0;
					while(j < l)
					{
						if(p[j] != '\0')
						{
							(void)log_msg(h, (sp + sizeof(struct ether_header)), (sp + sizeof(struct ether_header)), "Overlapping fragment");
							j = l;
						}
						j++;
					}
					memcpy((cp + o), (r->p + sizeof(struct ether_header) + hlen), l);
				}
			}
			memcpy((sp + rh), cp, (65536 - rh));
			ip2 = (struct ip *)(sp + sizeof(struct ether_header));
			ip2->ip_len = htons(tl + (rh - sizeof(struct ether_header)));
			ip2->ip_off = 0;
			ip2->ip_sum = 0;
			ip2->ip_sum = ip_checksum((u_char *)ip2, sizeof(struct ip));
			if((f1 - f0) != tl)
			{
				(void)log_msg(h, (sp +
					sizeof(struct ether_header)), (sp +
					sizeof(struct ether_header)),
					"Fragment reassembly problem");
				if(verbose > 0)
				{
					for(r = f->frags; r->next; r = r->next)
					{
						log_msg(r->h, (r->p +
							sizeof(struct
							ether_header)), (r->p
							+ sizeof(struct
							ether_header)),
							"Bad Frag");
					}
				}
			}
			else
			{
				h->len = ip->ip_len;
				if(bpf_filter(base_fcode.bf_insns, (char *)sp,
					h->len, h->caplen))
				{
					/*
					log_msg(NULL, NULL, NULL, "Foo Frag");
					*/
					pcap_print_pt("Reassembled frag", h,
						sp);
				}
			}
			free(cp);
			free(sp);
		}
	}
	free(h);
	return(0);
}

/* Stolen from one of the Stevens books.  -spb */
u_short
ip_checksum(u_char *p, int foo)
{
        long sum = 0;
        while(foo > 1)
        {
                sum += *((u_short *)p)++;
                if(sum & 0x80000000)
                        sum = (sum & 0xFFFF) + (sum >> 16);
                foo -= 2;
        }
        if(foo) sum += (u_short) *(u_char *)p;
        while(sum >> 16) sum = (sum & 0xFFFF) + (sum >> 16);
        return ~sum;
}

void
make_filter(int i, struct filter *filter)
{
	memcpy(&filter->filter, filter_pt[i], MAX_FILTER_LEN);
	memcpy(&filter->hex, hex_pt[i], MAX_HEX_LEN);
	memcpy(&filter->call_args, call_args_pt[i], MAX_CALLARG_LEN);
	memcpy(&filter->id, id_pt[i], MAX_ID_LEN);
	memcpy(&filter->type, type_pt[i], MAX_TYPE_LEN);
	filter->regex = regex_pt[i];
	filter->fcode = fcode_pt[i];
	filter->fcount = fcount_pt[i];
	filter->d = d_pt[i];
	filter->t = t_pt[i];
	filter->q = q_pt[i];
	filter->h = h_pt[i];
	filter->sev = sev_pt[i];
	filter->callback = callback_pt[i];
	filter->index = rn_pt[i];
}

void
doctrine_to_filter(struct doctrine *d, struct filter *f)
{
	memcpy(&f->filter, &d->filter, MAX_FILTER_LEN);
	memcpy(&f->hex, &d->hex, MAX_HEX_LEN);
	memcpy(&f->call_args, &d->call_args, MAX_CALLARG_LEN);
	memcpy(&f->id, &d->id, MAX_ID_LEN);
	memcpy(&f->type, &d->type, MAX_TYPE_LEN);
	f->regex = d->regex;
	f->fcode = d->fcode;
	f->fcount = d->fcount;
	f->d = d->d;
	f->t = d->t;
	f->q = d->q;
	f->h = d->h;
	f->sev = d->sev;
	f->callback = d->callback;
	f->index = d->index;
}
