#!/usr/bin/perl -T

use Pg;
use CGI;
use strict;

use vars qw/$DB $TABLE $selfref $selfurl %hidden_fields %label %table_label/;
use vars qw/$query $conn/;

$DB = 'foobar';
$TABLE = 'foobar';

$selfref = 'pgsql_query.cgi';
$selfurl = "https://meshuga.etimecapital.com/cgi-bin/$selfref";

%hidden_fields =
(
	'sql_query',	'',
	'view',		'',
);

%label = 
(
	'query',	'Run Query',
	'views',	'List Views',
	'generic',	'Generic Query',
	'show_view',	'Show View',
	'new_events',	'New Events',
	'submit',	'Submit',
	'clear_all',	'Clear All',
	'delete_all',	'Delete All',
);

%table_label = 
(
	'timestamp',	'Timestamp',
	'datetime',	'Date/Time',
	'dotsec',	'',
	'alert',	'Alert',
	'src',		'Source',
	'dst',		'Destination',
	'sport',	'Source Port',
	'dport',	'Dest. Port',
	'proto',	'Proto',
	'len',		'Length',
	'id',		'Event ID',
	'seq',		'TCP Seq',
	'ack',		'TCP Ack',
	'tos',		'TOS',
	'ttl',		'TTL',
	'ipid',		'IP ID',
	'win',		'Win',
	'icmp_type',	'ICMP Type',
	'flags',	'Flags',
	'comment',	'Comment',
	'domain',	'Domain',
	'whois',	'Whois',
	'trace',	'Trace',
	'dump',		'Dump',
	'hex',		'Hex',
	'sensor',	'Sensor',
	'level',	'Alert Lvl',
	'typeflag',	'Type',
	'newflag',	'New',
);


&main();
exit(0);

sub main
{
	&do_html_top();
	&clean_environment();
	&check_query();
	&do_html_bottom();
}

sub do_html_top
{
	my($foo);

	$query = new CGI;
	print $query->header;
	print $query->start_html("Query Interface");
	print $query->startform(-action=>$selfurl);
	foreach $foo (keys %hidden_fields)
	{
		$hidden_fields{$foo} = $query->param($foo);
		#print "<p>$foo:  $hidden_fields{$foo}\n";
	}
}

sub do_html_bottom
{
	my($foo, @bar);

	foreach $foo (sort(keys %hidden_fields))
	{
		$query->param($foo, $hidden_fields{$foo});
		print $query->hidden($foo, $hidden_fields{$foo});
		#print "<p>FOO hidden $foo:  $hidden_fields{$foo}<p>\n";
	}
	print "<p><hr><p>\n";
	print $query->submit('action', $label{'new_events'});
	print "&nbsp;&nbsp;";
	print $query->submit('action', $label{'views'});
	print "&nbsp;&nbsp;";
	print $query->submit('action', $label{'generic'});
	print $query->endform;
	print $query->end_html;
}

sub check_query
{
	my($foo, @p, $bar, $o, $q, $v, @format);

	if($query->param('action') eq $label{'new_events'})
	{
		&new_events();
		return;
	}
	elsif($query->param('action') eq $label{'views'})
	{
		&list_views();
		return;
	}
	elsif($query->param('action') eq $label{'generic'})
	{
		&generic_query();
		return;
	}
	if($query->param('sort_by'))
	{
		$foo = $query->param('sort_by');
		foreach $bar (keys %table_label)
		{
			($foo eq $table_label{$bar}) &&
				(($o = "order by $bar"),
				($hidden_fields{'sql_query'} =~
				s/(\s+)order(\s+)by(\s+)(\w+)(\s*)//g), last);
		}
	}
	$hidden_fields{'sql_query'} && ($hidden_fields{'sql_query'} .= " $o");
	$ENV{'QUERY_STRING'} && (&show_details($o), return);
	$q = $hidden_fields{'sql_query'};
	$v = $hidden_fields{'view'};
	$v || ($v = $TABLE);
	if($query->param('action') eq $label{'show_view'})
	{
		&show_view();
		return;
	}
	elsif($query->param('action') eq $label{'query'})
	{
		@format = &get_format($v);
		&parse_query($q, @format);
		return;
	}
	elsif($query->param('action') eq $label{'submit'})
	{
		&submit();
		return;
	}
	else
	{
		if($q)
		{
			unless($v)
			{
				$foo = $q;
				($foo =~ /from(\s+)(\w+)(\s+)/) ||
					(&main_menu(), return);
				$v = $2;
			}
			@format = &get_format($v);
			&parse_query($q, @format);
			return;
		}
		else
		{
			&main_menu();
			return;
		}
	}
}

sub main_menu
{
}

sub generic_query
{
	print "<p>Enter query:  \n";
	print $query->textfield(-name=>'sql_query',
		-value=>"select timestamp,hex from $TABLE",
		-size=>40, -maxlength=>160);
	print $query->submit('action', $label{'query'});
}

sub db_connect
{
	$conn = Pg::connectdb("dbname=$DB");
}

sub list_views
{
	my($q, $e, $res, @row, $foo, @views);

	&db_connect();
	$q = "select relname from pg_class where relname \!\~ \'\^pg_\' and relkind=\'r\' and relhasrules order by relname";
	$foo = $conn->options;
	#print "Options:  ($foo)\n";
	$res = $conn->exec($q);
	while(@row = $res->fetchrow)
	{
		$foo = join(" ", @row);
		push(@views, $foo);
	}
	$e = $conn->errorMessage;
	$e && (print "<hr><b>Error:</b><br>\n$e\n<hr>\n");
	print "<p><b>Select a View</b><p>\n";
	print $query->popup_menu(-name=>'selected_view', -values=>\@views,
		-size=>10);
	print "<br>\n";
	print $query->submit('action', $label{'show_view'});
	print "<br>\n";
}

sub show_view
{
	my($view, @format, $q);
	$view = join(' ', $query->param('selected_view'));
	@format = &get_format($view);
	$q = "select * from $view";
	$hidden_fields{'sql_query'} = $q;
	$hidden_fields{'view'} = $view;
	&parse_query($q, @format);
}

sub show_details
{
	my($o) = @_;
	my(@q, $q2, $view, @format, $foo, @bar, $baz);

	$view = $hidden_fields{'view'};
	($view eq '') && ($view = 'new_events');
	@q = split('&', $ENV{'QUERY_STRING'});
	foreach $foo (@q)
	{
		@bar = split('=', $foo);
		($#bar == 1) || next;
		($bar[0] eq 'view') && (($view = $bar[1]), next);
		$baz .= "($bar[0] = \'$bar[1]\')";
	}
	$q2 = "select * from $view where $baz $o";
	@format = &get_format($view);
	$hidden_fields{'sql_query'} = $q2;
	$hidden_fields{'view'} = $view;
	&parse_query($q2, @format);
}

sub parse_query
{
	my($q, @format) = @_;
	my($line, $foo, $res, @row, $e);
	print "<h3 align=center>$q</h3>\n";
	print "<br><table border=1 align=center>\n";
	&db_connect();
	#$q =~ s/select(\s+)//g;
	$res = $conn->exec($q);
	print "<tr>";
	print "<td>Action</td>";
	foreach $foo (@format)
	{
		#print "<td>$table_label{$foo}</td>";
		($foo eq 'oid') && next;
		print "<td>";
		$table_label{$foo} &&
			(print $query->submit('sort_by', $table_label{$foo}));
		print "</td>";
	}
	print "</tr>\n";
	while(@row = $res->fetchrow)
	{
		$foo = join("##x#", @row);
		&parse_row($foo, @format);
	}
	print "</table>";
	$e = $conn->errorMessage;
	$e && (print "<hr><b>Error:</b><br>\n$e\n<hr>\n");
	print $query->popup_menu('Global', ['', 'Clear All', 'Delete All'], '');
	print $query->submit('action', $label{'submit'});
}

sub parse_row
{
	my($line, @format) = @_;
	my($foo, @bar, $baz, $blah, $i, $v, $oid);

	$v = $hidden_fields{'view'};
	$foo = join(' ', @format);
	@bar = split('##x#', $line);
	$baz = "<a href=$selfref";
	$i = -1;
	print "<tr>\n";
	print "<td>";
	print $query->popup_menu('Foo', [$oid, 'Clear', 'Delete'], '');
	print "</td>";
	SWITCH: for (@format)
	{
		$i++;
		print "<td>";
		$blah = "$bar[$i]&view=$v>$bar[$i]</a></td>";
		/datetime/	&& do { print "$bar[$i]</td>"; next; };
		/timestamp/	&& do { print "$baz?timestamp=$blah"; next; };
		/dotsec/	&& do { print "$bar[$i]</td>"; next; };
		/alert/		&& do { print "$baz?alert=$blah"; next; };
		/src/		&& do { print "$baz?src=$blah"; next; };
		/sport/		&& do { print "$baz?sport=$blah"; next; };
		/dst/		&& do { print "$baz?dst=$blah"; next; };
		/dport/		&& do { print "$baz?dport=$blah"; next; };
		/proto/		&& do { print "$baz?proto=$blah"; next; };
		/len/		&& do { print "$bar[$i]</td>"; next; };
		/id/		&& do { print "$bar[$i]</td>"; next; };
		/seq/		&& do { print "$bar[$i]</td>"; next; };
		/ack/		&& do { print "$bar[$i]</td>"; next; };
		/tos/		&& do { print "$bar[$i]</td>"; next; };
		/ttl/		&& do { print "$bar[$i]</td>"; next; };
		/ipid/		&& do { print "$bar[$i]</td>"; next; };
		/win/		&& do { print "$bar[$i]</td>"; next; };
		/icmp_type/	&& do { print "$bar[$i]</td>"; next; };
		/flags/		&& do { print "$bar[$i]</td>"; next; };
		/comment/	&& do { print "$bar[$i]</td>"; next; };
		/domain/	&& do { print "$bar[$i]</td>"; next; };
		/whois/		&& do { print "$bar[$i]</td>"; next; };
		/trace/		&& do { print "$bar[$i]</td>"; next; };
		/dump/		&& do { print "$bar[$i]</td>"; next; };
		/hex/		&& do { print "$bar[$i]</td>"; next; };
		/sensor/	&& do { print "$bar[$i]</td>"; next; };
		/level/		&& do { print "$bar[$i]</td>"; next; };
		/typeflag/	&& do { print "$bar[$i]</td>"; next; };
		/newflag/	&& do { print "$bar[$i]</td>"; next; };
		print "N/A";
	}
	print "</tr>\n";
}


sub new_events
{
	my($stuff, @format, $q);

	$stuff = 'datetime,timestamp,dotsec,alert,src,sport,dst,dport,proto';
	@format = split(',', $stuff);
	$q = "select $stuff from new_events";
	$hidden_fields{'sql_query'} = $q;
	$hidden_fields{'view'} = 'new_events';
	&parse_query($q, @format);
}

sub get_format
{
	my($view) = @_;
	my($q, $res, $foo, @bar, $e, @row, @format);

	$q = "select attname from pg_attribute where (pg_class.relname=\'$view\') and (pg_class.relowner=pg_user.usesysid) and (pg_class.oid=pg_attribute.attrelid) and (pg_attribute.atttypid=pg_type.oid) and (attnum > 0)";
	&db_connect();
	$res = $conn->exec($q);
	while(@row = $res->fetchrow)
	{
		$foo = join(' ', @row);
		unshift(@format, $foo);
	}
	$e = $conn->errorMessage;
	$e && (print "<hr><b>Error:</b><br>\n$e\n<hr>\n");
	return(@format);
}

sub show_full
{
	my(@format, $q);

	@format = keys %table_label;
}

sub clean_environment
{
	my($foo, $bar);
	foreach $foo (sort(keys %ENV))
	{
		($foo eq 'REMOTE_USER') &&
			(($ENV{$foo} =~ tr/a-zA-Z0-9//cd), next);
		($foo eq 'QUERY_STRING') &&
			(($ENV{$foo} =~ tr/a-zA-Z0-9=\.&_//cd), next);
		delete $ENV{$foo};
	}
}

sub submit
{
	my($q, $a);
	$q = $hidden_fields{'sql_query'};
	$a = $query->param('Global');
	print "<p>$a<p>\n";
	if($a eq 'Delete All')
	{
		$q =~ s/select/delete/g;
		print "<p>Run:  $q<p>\n";
	}
	elsif($a eq 'Clear All')
	{
		$q =~ s/select/update/g;
		$q .= " set newflag=0";
		print "<p>Run:  $q<p>\n";
	}
	else
	{
	}
}
