#!/usr/bin/perl
#
# This is a simple script to convert snort signatures into shoki filter
# rules.  Specify the input file with the -f flag.  The converted shoki
# filters are written to stdout.
#
# Currently, the only filters this script tries to convert are those
# with `content:' strings.  If multiple search expressions occur in
# a single `content:' string, you'll have to edit the resulting filter
# rule by hand---shoki only understands a single search expression (although
# you can use POSIX extended regexen).
#

%tcp_flags =
(
	'F',    '(tcp[13]&1!=0)',
	'S',    '(tcp[13]&2!=0)',
	'R',    '(tcp[13]&4!=0)',
	'P',    '(tcp[13]&8!=0)',
	'A',    '(tcp[13]&10!=0)',
	'U',    '(tcp[13]&20!=0)',
	'B',    '(tcp[13]&40!=0)',
);
require("getopts.pl");
&Getopts("f:");
$file = $opt_f;
$file || (die "Specify an input file using -f\n");

&main();
exit(0);


sub main
{
	&read_file($file);
}

sub read_file
{
	my($fname) = @_;
	my($proto, $search, $msg, $id, $comment);
	open(FILE, "<$fname") || (die "Couldn't read $fname\n");
	while(<FILE>)
	{
		$proto = '';
		$id = '';
		$search = '';
		$msg = '';
		$comment = '';
		s/^alert //g || next;
		s/^udp //g && ($proto = 'udp');
		s/^tcp //g && ($proto = 'tcp');
		s/^icmp //g && ($proto = 'icmp');
		s/^UDP //g && ($proto = 'udp');
		s/^TCP //g && ($proto = 'tcp');
		s/^ICMP //g && ($proto = 'icmp');
		$proto || next;
		/content:(\s*)"(.+)"/ || next;
		$search = $2;
		$search = &convert_search($search);
		/msg:(\s*)"(.+)"/ || next;
		$msg = $2;
		($msg =~ /"(\s*);/) && ($msg = $`);
		if($msg =~ /\//)
		{
			($id, $comment) = split('\/', $msg);
		}
		else
		{
			$id = 'NULL';
			$comment = $msg;
		}
		/(\w+)(\s*)->(\s*).(\w+)(\s+)(\w+)/ || next;
		$sport = $1;
		$dport = $6;
		$flags = '';
		/flags:(\s*)(\w+);/ && ($flags = &convert_flags($2));
		$filter = "$proto";
		($sport =~ /^(\d+)$/) && ($filter .= " and src port $sport");
		($dport =~ /^(\d+)$/) && ($filter .= " and dst port $dport");
		$flags && ($filter .= " $flags");
		($id =~ /^(\s*)$/) && ($id = 'NULL');
		print "$filter\t65536\tSEARCH\t$id\t$comment\t$search\tALL\n";
	}

}

sub convert_search
{
	my($foo) = @_;
	my(@bar, $baz, $blah);
	($foo =~ /"(\s*);/) && ($foo = $`);
	# alternation omitted, as it is used by snort rules to
	# delimit hex literals
	$foo =~ s/(?=[.\[\]$()*+?{}])/\\/g;
	while($foo =~ /\|(.+?)\|/)
	{
		$bar = $1;
		$blah = $bar;
		$bar =~ s/(\s+)//g;
		$foo =~ s/\|$blah\|/0x$bar/;
	}
	return($foo);
}

sub convert_flags
{
	my($foo) = @_;
	my($bar);
	$foo =~ s/(\d+)//g;
	foreach $bar (keys %tcp_flags)
	{
		$foo =~ s/$bar/ and $tcp_flags{$bar}/g;
	}
	return($foo);
}
