#!/usr/bin/perl
#
# This script uses the perl GD module(s) to generate plots.
#
#	-a	Additional flags to pass to the widget.  Use -W to
#		define the widget.
#	-e	A pcap filter expression
#	-E	End time
#	-i	Number of horizontal increments
#	-I	Force increment to equal specified number of seconds
#	-l	Make the vertical axis logarithmic
#	-r	Either a single dump file, or a list of them.  If a list,
#		it must be double quoted.
#	-s	Name of the sensor the data is coming from
#	-S	Start time
#	-W	Define which widget to use
#
# The arguments to -E and -S should be in seconds after the start of the
# epoch.  Only packets stamped after -S and before -E are used.  If these
# endpoints don't fall withing the period of time covered by the specified
# logfiles, you'll get a lopsided graph.  I.e., if the start time of
# the data is after -S, the span between the two will be blank in the
# graph.
#

&read_config("/etc/shoki.conf");
require("$INSTALL_PREFIX/$ETCDIR/collector.conf");
$logdir = "$INSTALL_PREFIX/central/log";
$plotdir = "$INSTALL_PREFIX/plot";
#$plotdir = "/usr/local/vmware/ids/plot";
$service = "$INSTALL_PREFIX/$BINDIR/service";

use GD::Graph::lines;

require("getopts.pl");
&Getopts("a:e:E:i:I:ln:r:s:S:t:W:");
&main();
exit(0);

sub main
{
	&parse_input();
	&run_service();
	&make_plot();
}


sub parse_input
{
	my($foo, @bar, $baz, $blah);
	$opt_i || ($opt_i = 15);
	#$opt_s || (die "Specify a damn sensor.\n");
	$opt_r || (die "Specify a damn log file.\n");
	$opt_W && ($widget = $opt_W) || ($widget = $service);
	$opt_a && ($args = $opt_a) || ($args = "-p -t -q -P 5");
	@dumps = split('\s+', $opt_r);
	foreach $dump (@dumps)
	{
		(-f $dump) || (die "File $dump doesn't exist.  Bailing.\n");
		@bar = split('\/', $dump);
		($bar[$#bar] =~ /\.(\d+)-(\d+)\./)
			|| (die "Filename $dump is invalid.\n");
		(!$first || ($1 < $first)) && ($first = $1);
		(!$last || ($2 > $last)) && ($last = $2);
		$dumplist .= "$dump ";
	}
	$last -= 5;
	$opt_S && ($first = $opt_S);
	$opt_E && ($last = $opt_E);
	$baz = $opt_i;
	$blah = 0;
	while(($opt_i != 0) && ($blah < ($last - $first)))
	{
		$blah++;
		$opt_i = $baz + $blah;
		((($last - $first) % $opt_i) == 0) && last;
		$opt_i = $baz - $blah;
		($opt_i < 1) && next;
		((($last - $first) % $opt_i) == 0) && last;
	}
	$inc = ($last - $first) / $opt_i;
	$opt_I && ($inc = int($opt_I));
}

sub run_service
{
	my($foo, $bar, $baz, $blah);
	my($port, $count);
	$foo = $first;
	@x = ();
	#@xlabel = ();
	@ports = ();
	while($foo <= ($last - $inc))
	{
		$bar = $foo + $inc;
		$baz = int(($foo + $bar) / 2.0);
		push(@x, int($baz));
		$foo += $inc;
	}
	$foo = $first;
	#print "$widget -r \"$dumplist\" -S $first -E $bar -I $inc $args $opt_e\n";
	open(PIPE, "$widget -r \"$dumplist\" -S $first -E $last -I $inc $args $opt_e|") ||
		(die "Problem with $widget pipe.\n");
	while(<PIPE>)
	{
		chomp;
		s/^(\s+)//g;
		s/(\s+)$//g;
		($port, $count) = split('\s+', $_);
		#$port = int($port);
		$count = int($count);
		($port && $count) || (($bar = $foo + $inc),
			($baz = int(($foo + $bar) / 2.0), ($foo += $inc)),
			next);
		#${"port$port_$baz"} = int($count);
		($opt_l && ($portcount{$port}{$baz} = log($count)))
			|| ($portcount{$port}{$baz} = int($count));
		$portsa{$port}++;
	}
	close(PIPE);
	foreach $foo (sort(keys %portsa))
	{
		push(@ports, $foo);
	}
	push(@data, \@x);
	foreach $foo (@ports)
	{
		undef(@{$foo});
		foreach $bar (@x)
		{
			push(@{$foo}, ($portcount{$foo}{$bar} / $inc));
			#print int($portcount{$foo}{$bar}) . " ";
		}
		#print "\n$foo:  $#d\n";
		push(@data, \@{$foo});
	}
}

sub make_plot
{
	my($graph, $title, $ffirst, $flast, $i);
	$title = "Traffic from " . localtime($first) . " to " .
		localtime($last);
	$graph = GD::Graph::lines->new(600,450);
	$ffirst = $first;
	$flast = $last;
	#while(($ffirst % $inc) != 0)
	#{
		#$ffirst--;
	#}
	#while(($flast % $inc) != 0)
	#{
		#$flast++;
	#}
	#print "$first - $last (" . ($last - $first) . ")\n";
	#print "f:  $ffirst - $flast (" . ($flast - $ffirst) . ")\n";
	$tick_count = int(($last - $first) / $inc);
	$i = 0;
	$j = $tick_count;
	#while(($tick_count > 20) && ((($last - $first) % $tick_count) != 0))
	#{
	#}
	$opt_t && ($tick_count = $opt_t);
	#print "tick_count:  $tick_count\n";
	$graph->set(x_label=>'Time', y_label=>'Bytes/sec', bgclr=>'lgray',
		fgclr=>'dgray', transparent=>0, line_types => [1,2,3,4],
		x_labels_vertical=>1, long_ticks=>1, line_width=>2,
		x_tick_number=>$tick_count, title=>$title,
		x_min_value=>$ffirst, x_max_value=>$flast);
	$graph->set_legend(@ports);
	$graph->plot(\@data);
	&save_graph($graph, $opt_n);
}

sub read_config
{
	my($fname) = @_;
	open(FILE, "<$fname") ||
		(die "Couldn't read $fname");
	while(<FILE>)
	{
		/^(\w+)(\s*)=(\s*)(.+)$/ || next;
		${$1} = $4;
	}
	close(FILE);
}


sub save_graph
{
	my($graph, $name) = @_;
	my($ext);
	$ext = $graph->export_format;
	open(OUT, ">$plotdir/$name.$ext") ||
		(die "Couldn't write $plotdir/$name.$ext");
	binmode OUT;
	print OUT $graph->gd->$ext();
	close(OUT);
}
