#include "shoki_includes.h"
#include "shoki.h"
#include "shoki_init.h"
#include "shoki_doctrine.h"
#include "shoki_config.h"
#include "shoki_callback.h"
#include "shoki_stub.h"

#define SWAPLONG(y) \
((((y)&0xff)<<24) | (((y)&0xff00)<<8) | (((y)&0xff0000)>>8) | (((y)>>24)&0xff))

#define TCPDUMP_MAGIC 0xa1b2c3d4


int default_snaplen, start, stop, firstsec, lastsec, silent = 0;
char *currentfile, *dev, *outfile, *fname, *filterlist;
int nfilters, fcount_pt[MAX_FILTERS], callback_pt[MAX_FILTERS];
int d_pt[MAX_FILTERS], t_pt[MAX_FILTERS], q_pt[MAX_FILTERS];
int h_pt[MAX_FILTERS];
char **filter_pt, **call_args_pt, **hex_pt, **type_pt, **id_pt;
regex_t **regex_pt;
char *dumpfile = NULL;
struct bpf_program **fcode_pt;
struct bpf_program base_fcode;
int callbacks = 1;
int linux_kludge = 0, timeval_kludge = 0, fragre = 0;
int quick = 0;

FILE *logfile = NULL;
int fd;
pcap_t *pd;

int pf_outformat[] = {};
char *pf_outdelim[] = {};

pcap_handler out = pcap_print_pt;


int
main(int argc, char **argv)
{
	register int op;
	register char *cfile = NULL, *config = NULL;

	shoki_type = strdup(SHOKI_TYPE);
	shoki_log_name = strdup(SHOKI_LOG_NAME);

	while((op = getopt(argc, argv, "C:E:F:kKL:r:s:S:")) != EOF)
		switch(op)
		{
			case 'C':
				cfile = optarg;
				break;
			case 'E':
				stop = atoi(optarg);
				break;
			case 'F':
				filterlist = optarg;
				break;
			case 'k':
				linux_kludge = 1;
				break;
			case 'K':
				timeval_kludge = 1;
				break;
			case 'L':
				currentfile = optarg;
				break;
			case 'r':
				dumpfile = optarg;
				break;
			case 's':
				default_snaplen = atoi(optarg);
				break;
			case 'S':
				start = atoi(optarg);	
				break;
		}

	base_filter = get_filter(&argv[optind]);

	((cfile == NULL) ? (parse_shoki_config(MATCH_CONFIG)) :
		(parse_shoki_config(cfile)));
	if((fname != NULL) && (base_filter == NULL))
		base_filter = read_file(fname);
	if(filterlist == NULL) error("No filterlist defined.");
	config = read_file(filterlist);
	if(dumpfile == NULL) error("No dumpfile");
	if(filterlist != NULL) (void)parse_config(filterlist);
	if(default_snaplen == -1) (default_snaplen = DEFAULT_SNAPLEN);
	(void)fake_pcap(pd, 0);
	if(currentfile != NULL)
	{
		if(*currentfile == '-') logfile = stdout;
		else logfile = fopen(currentfile, "a");
	}
	if(logfile == NULL) (void)log_msg(NULL, NULL, NULL, "stub starting");
	(void)read_zdumps(dumpfile, out);
	if(logfile) (void)fclose(logfile);
	exit(0);
}

void
pcap_print_pt(u_char *user, const struct pcap_pkthdr *h, const u_char *sp)
{
	int foo, i;
	if(start != 0) if(h->ts.tv_sec < start) return;
	if(stop != 0) if(h->ts.tv_sec > stop) return;
	foo = 0;
	if(nfilters == 0) error("No filters defined");
	for(i = 1; i <= nfilters && foo == 0; i++)
	{
		if(bpf_filter(fcode_pt[i]->bf_insns, (u_char *)sp, h->len, h->caplen))
		{
			foo = 1;
		}
	}
}
